# -*- coding: utf-8 -*-

from textwrap import indent


SERDE_FORMAT = "org.openx.data.jsonserde.JsonSerDe"
INPUTFORMAT = "org.apache.hadoop.mapred.TextInputFormat"
OUTPUTFORMAT = "org.apache.hadoop.hive.ql.io.IgnoreKeyTextOutputFormat"


def indent_quoted(s, n=1):
    spc = " " * n * 4
    return indent(f"'{s}'", spc)


def create_table(definitions, schema=None, table=None):
    """Construct `create table` statement in DDL.

    Arg:
        definitions (str):
            - field name : dtype map
    Kwargs:
        schema (str):
            - default: None
            - schema to template
        table (str):
            - default: None
            - table name to template

    Returns: str
    """

    if not schema:
        schema = "{schema}"
    if not table:
        table = "{table}"

    create = f"CREATE EXTERNAL TABLE {schema}.{table}"

    definitions = definitions.strip("\n")
    if not definitions.startswith(" "):
        definitions = indent(definitions, "    ")
    return f"{create} (\n{definitions}\n)"


def format_s3_key(s3_key):
    if s3_key:
        s3_key = s3_key.strip()
        if not s3_key.startswith("s3://"):
            s3_key = f"s3://{s3_key}"
    else:
        s3_key = "s3://{bucket}/{prefix}"
    return s3_key


def format_partitions(partitions):
    partition_by = ""
    if partitions:
        key_types = (f"{key} {dtype.upper()}" for (key, dtype) in partitions.items())
        partition_by = f"PARTITIONED BY ({', '.join(key_types)})"
    return partition_by


def set_options(
    key_map=None,
    partitions=None,
    s3_key=None,
    case_insensitive=True,
    ignore_malformed_json=True,
):
    """Construct options statement in DDL.

    Kwargs:
        key_map (dict):
            - default: None
            - mapping for for column names
        partitions (dict):
            - default: None
        s3_key (str):
            - default: None
            - S3 key prefix
        case_insensitive (bool):
            - default: True
            - set all fields to lower case
        ignore_malformed_json (bool):
            - default: True
            - instruct Spectrum to ignore bad JSON in data lake

    Returns: str
    """

    bool_str = lambda b: "TRUE" if b else "FALSE"

    # OpenX Serde Properties
    serde_properties = []
    if key_map:
        mappings = [f"'mapping.{key}'='{val}'" for key, val in key_map.items()]
        serde_properties.extend(mappings)

    case_option = f"'case.insensitive'='{bool_str(case_insensitive)}'"
    serde_properties.append(case_option)
    json_option = f"'ignore.malformed.json'='{bool_str(ignore_malformed_json)}'"
    serde_properties.append(json_option)
    serde_properties = ",\n".join(serde_properties)

    statement = f"""
{format_partitions(partitions)}
ROW FORMAT SERDE
{indent_quoted(SERDE_FORMAT)}
WITH SERDEPROPERTIES (
{indent(serde_properties, '    ')}
)
STORED AS INPUTFORMAT
{indent_quoted(INPUTFORMAT)}
OUTPUTFORMAT
{indent_quoted(OUTPUTFORMAT)}
LOCATION '{format_s3_key(s3_key)}';"""

    return statement.strip()


def create_statement(
    definitions,
    key_map=None,
    schema=None,
    table=None,
    partitions=None,
    s3_key=None,
    case_insensitive=True,
    ignore_malformed_json=True,
):
    """Construct Spectrum DDL.

    Arg:
        definitions (str):
            - field name : dtype map
    Kwargs:
        schema (str):
            - default: None
            - schema to template
        table (str):
            - default: None
            - table name to template
        key_map (dict):
            - default: None
            - mapping for for column names
        partitions (dict):
            - default: None
            - {column: dtype}
        s3_key (str):
            - default: None
            - S3 key prefix
        case_insensitive (bool):
            - default: True
            - set all fields to lower case
        ignore_malformed_json (bool):
            - default: True
            - instruct Spectrum to ignore bad JSON in data lake

    Returns: str
    """

    create_external_table = create_table(definitions, schema, table)

    options = set_options(
        key_map=key_map,
        partitions=partitions,
        s3_key=s3_key,
        case_insensitive=case_insensitive,
        ignore_malformed_json=ignore_malformed_json,
    )

    return "\n".join((create_external_table, options))



# -*- coding: utf-8 -*-

from enum import IntEnum, auto
from functools import singledispatch

from . import parse_date


REDSHIFT_MAP = {
    "int2": "SMALLINT",
    "smallint": "SMALLINT",
    "int": "INT",
    "int4": "INT",
    "integer": "INT",
    "int8": "BIGINT",
    "bigint": "BIGINT",
    "real": "FLOAT4",
    "float": "FLOAT8",
    "double precision": "FLOAT8",
    "boolean": "BOOL",
    "character": "VARCHAR",
    "character varying": "VARCHAR",
    "date": "DATE",
    "timestamp without time zone": "TIMESTAMP",
}

REDSHIFT_ALIAS = set(REDSHIFT_MAP.values())


class Numeric(IntEnum):
    SMALLINT = auto()
    INT = auto()
    BIGINT = auto()
    FLOAT4 = auto()
    FLOAT8 = auto()


DATETIME_TYPES = {"DATE", "TIMESTAMP"}
NUMERIC_TYPES = set(Numeric.__members__.keys())


def largest_numeric_type(*dtypes: str) -> str:
    """Get dtype with largest range."""

    num_type = max((getattr(Numeric, dtype) for dtype in dtypes))
    return num_type.name


def resolve_type(inferred_dtype: str, user_dtype: str) -> str:
    """Detect and resolve data type when inferred != user provided.

    Rules (inferred vs. user defined):
        string / varchar > all:
            - override all dtypes except for date/time
        numeric > all numeric:
            - user defined numeric(x,y) types override int, float
            - other inferred dtypes override user defined
        int < bool:
            - user defined bool replaces inferred int
        int < float:
            - user defined float is used since spectrum supports int to float
        int <-> int:
            - largest int type is used
        float <-> float:
            - largest float type is used
    """

    dtype = None

    if inferred_dtype == "VARCHAR":
        if user_dtype in DATETIME_TYPES:
            dtype = user_dtype
        else:
            dtype = inferred_dtype
    elif "NUMERIC" in user_dtype or "DECIMAL" in user_dtype:
        if inferred_dtype in NUMERIC_TYPES:
            dtype = user_dtype
        else:
            dtype = inferred_dtype
    elif "INT" in inferred_dtype:
        if user_dtype == "BOOL" or "FLOAT" in user_dtype:
            dtype = user_dtype
        elif "INT" in user_dtype:
            dtype = largest_numeric_type(inferred_dtype, user_dtype)
        else:
            dtype = inferred_dtype
    elif "FLOAT" in inferred_dtype:
        if "FLOAT" in user_dtype:
            dtype = largest_numeric_type(inferred_dtype, user_dtype)
        else:
            dtype = inferred_dtype
    else:
        dtype = user_dtype

    return dtype


def type_set(t: list):
    return set(map(type, t))


def int_in_bounds(num_bits, val):
    n_max = (2 ** num_bits) // 2
    if val > 0:
        n_max -= 1
    return abs(val) <= n_max


@singledispatch
def set_dtype(val, **kwargs):
    return f"UNKNOWN_{type(val).__name__}"


@set_dtype.register
def __str_dtype(val: str, infer_date: bool = False):

    if any(c.isdigit() for c in val):
        dtype = parse_date.guess_type(val)
        if dtype:
            return dtype
    return "VARCHAR"


@set_dtype.register
def __float_dtype(val: float, **kwargs):

    num_bits = (64 - 15, 32 - 6)
    dtypes = ("FLOAT8", "FLOAT4")

    dtype = None
    for n, dtype in zip(num_bits, dtypes):
        if abs(val) >= 2 ** n // 2:
            break
    return dtype


@set_dtype.register
def __int_dtype(val: int, **kwargs):

    if int_in_bounds(16, val):
        dtype = "SMALLINT"
    elif int_in_bounds(32, val):
        dtype = "INT"
    elif int_in_bounds(64, val):
        dtype = "BIGINT"
    else:
        raise ValueError(f"Input exceeds integer max number of bits: {val}")
    return dtype


@set_dtype.register
def __bool_dtype(val: bool, **kwargs):
    return "BOOL"



# -*- coding: utf-8 -*-

from pendulum import parse as strict_parse
from pendulum.exceptions import ParserError


def num_digits(s: str):
    return sum(c.isdigit() for c in s)


def guess_type(s: str):
    """Guess datetime data type for input string. Supports ISO 8601.
    Uses pendulum parser to determine if input is valid date(time).
    """

    n = num_digits(s)

    # ISO 8601 dates have at least 8 digits
    if n < 8:
        return None

    try:
        strict_parse(s)
    except ValueError:
        return None
    except ParserError:
        return None

    if n == 8:
        dtype = "DATE"
    else:
        dtype = "TIMESTAMP"
    return dtype



# -*- coding: utf-8 -*-

keywords = {
    "abbrev",
    "abs",
    "abstime",
    "abstimeeq",
    "abstimege",
    "abstimegt",
    "abstimein",
    "abstimele",
    "abstimelt",
    "abstimene",
    "abstimeout",
    "abstimerecv",
    "abstimesend",
    "aclcontains",
    "aclinsert",
    "aclitemeq",
    "aclitemin",
    "aclitemout",
    "aclremove",
    "acos",
    "add_months",
    "aes128",
    "aes256",
    "age",
    "all",
    "alloc_rowid",
    "allowoverwrite",
    "alt_to_iso",
    "alt_to_koi8r",
    "alt_to_mic",
    "alt_to_win1251",
    "analyse",
    "analyze",
    "and",
    "any",
    "any_in",
    "any_out",
    "anyarray_in",
    "anyarray_out",
    "anyarray_recv",
    "anyarray_send",
    "anyelement_in",
    "anyelement_out",
    "area",
    "areajoinsel",
    "areasel",
    "array",
    "array_append",
    "array_cat",
    "array_coerce",
    "array_dims",
    "array_eq",
    "array_ge",
    "array_gt",
    "array_in",
    "array_le",
    "array_length_coerce",
    "array_lower",
    "array_lt",
    "array_ne",
    "array_out",
    "array_prepend",
    "array_recv",
    "array_send",
    "array_type_length_coerce",
    "array_upper",
    "as",
    "asc",
    "ascii",
    "ascii_to_mic",
    "asin",
    "atan",
    "atan2",
    "authorization",
    "avg",
    "az64",
    "backup",
    "between",
    "big5_to_euc_tw",
    "big5_to_mic",
    "binary",
    "bit",
    "bit_and",
    "bit_in",
    "bit_length",
    "bit_or",
    "bit_out",
    "bit_recv",
    "bit_send",
    "bitand",
    "bitcat",
    "bitcmp",
    "biteq",
    "bitge",
    "bitgt",
    "bitle",
    "bitlt",
    "bitne",
    "bitnot",
    "bitor",
    "bitxor",
    "blanksasnull",
    "bool",
    "bool_and",
    "bool_or",
    "booleq",
    "boolge",
    "boolgt",
    "boolin",
    "boolle",
    "boollt",
    "boolne",
    "boolout",
    "boolrecv",
    "boolsend",
    "both",
    "box",
    "box_above",
    "box_add",
    "box_below",
    "box_center",
    "box_contain",
    "box_contained",
    "box_distance",
    "box_div",
    "box_eq",
    "box_ge",
    "box_gt",
    "box_in",
    "box_intersect",
    "box_le",
    "box_lt",
    "box_mul",
    "box_out",
    "box_overlap",
    "box_recv",
    "box_same",
    "box_send",
    "box_sub",
    "bpchar",
    "bpchar_pattern_eq",
    "bpchar_pattern_ge",
    "bpchar_pattern_gt",
    "bpchar_pattern_le",
    "bpchar_pattern_lt",
    "bpchar_pattern_ne",
    "bpcharcmp",
    "bpchareq",
    "bpcharge",
    "bpchargt",
    "bpcharle",
    "bpcharlike",
    "bpcharlt",
    "bpcharne",
    "bpcharnlike",
    "bpcharout",
    "bpcharrecv",
    "bpcharregexeq",
    "bpcharregexne",
    "bpcharsend",
    "broadcast",
    "btabstimecmp",
    "btarraycmp",
    "btbeginscan",
    "btboolcmp",
    "btbpchar_pattern_cmp",
    "btbuild",
    "btbulkdelete",
    "btcharcmp",
    "btcostestimate",
    "btendscan",
    "btgettuple",
    "btinsert",
    "btint24cmp",
    "btint28cmp",
    "btint2cmp",
    "btint42cmp",
    "btint48cmp",
    "btint4cmp",
    "btint82cmp",
    "btint84cmp",
    "btint8cmp",
    "btmarkpos",
    "btname_pattern_cmp",
    "btnamecmp",
    "btoidcmp",
    "btoidvectorcmp",
    "btreltimecmp",
    "btrescan",
    "btrestrpos",
    "bttext_pattern_cmp",
    "bttextcmp",
    "bttintervalcmp",
    "btvacuumcleanup",
    "byteacat",
    "byteacmp",
    "byteaeq",
    "byteage",
    "byteagt",
    "byteain",
    "byteale",
    "bytealike",
    "bytealt",
    "byteane",
    "byteanlike",
    "byteaout",
    "bytearecv",
    "byteasend",
    "bytedict",
    "bzip2",
    "card_check",
    "case",
    "cash_cmp",
    "cash_div_int2",
    "cash_div_int4",
    "cash_eq",
    "cash_ge",
    "cash_gt",
    "cash_in",
    "cash_le",
    "cash_lt",
    "cash_mi",
    "cash_mul_int2",
    "cash_mul_int4",
    "cash_ne",
    "cash_out",
    "cash_pl",
    "cash_recv",
    "cash_send",
    "cash_words",
    "cashlarger",
    "cashsmaller",
    "cast",
    "cbrt",
    "ceil",
    "ceiling",
    "center",
    "char",
    "char_length",
    "character_length",
    "chareq",
    "charge",
    "chargt",
    "charle",
    "charlt",
    "charne",
    "charout",
    "charrecv",
    "charsend",
    "check",
    "checksum",
    "chr",
    "cideq",
    "cidin",
    "cidout",
    "cidr",
    "cidr_in",
    "cidr_out",
    "cidr_recv",
    "cidr_send",
    "cidrecv",
    "cidsend",
    "circle",
    "circle_above",
    "circle_add_pt",
    "circle_below",
    "circle_center",
    "circle_contain",
    "circle_contain_pt",
    "circle_contained",
    "circle_distance",
    "circle_div_pt",
    "circle_eq",
    "circle_ge",
    "circle_gt",
    "circle_in",
    "circle_le",
    "circle_lt",
    "circle_mul_pt",
    "circle_ne",
    "circle_out",
    "circle_overlap",
    "circle_recv",
    "circle_same",
    "circle_send",
    "circle_sub_pt",
    "close_lb",
    "close_ls",
    "close_lseg",
    "close_pb",
    "close_pl",
    "close_ps",
    "close_sb",
    "close_sl",
    "collate",
    "column",
    "concat",
    "constraint",
    "contains",
    "contjoinsel",
    "contsel",
    "convert",
    "convert_timezone",
    "convert_using",
    "cos",
    "cot",
    "count",
    "crc32",
    "create",
    "credentials",
    "cross",
    "cume_dist",
    "current_database",
    "current_date",
    "current_schema",
    "current_schemas",
    "current_setting",
    "current_time",
    "current_timestamp",
    "current_user",
    "current_user_id",
    "currtid",
    "currtid2",
    "currval",
    "cvtile_sf",
    "date",
    "date_add",
    "date_cmp",
    "date_cmp_timestamp",
    "date_cmp_timestamptz",
    "date_eq",
    "date_eq_timestamp",
    "date_eq_timestamptz",
    "date_ge",
    "date_ge_timestamp",
    "date_ge_timestamptz",
    "date_gt",
    "date_gt_timestamp",
    "date_gt_timestamptz",
    "date_in",
    "date_larger",
    "date_le",
    "date_le_timestamp",
    "date_le_timestamptz",
    "date_lt",
    "date_lt_timestamp",
    "date_lt_timestamptz",
    "date_mi",
    "date_mi_interval",
    "date_mii",
    "date_ne",
    "date_ne_timestamp",
    "date_ne_timestamptz",
    "date_out",
    "date_part",
    "date_part_year",
    "date_pl_interval",
    "date_pli",
    "date_recv",
    "date_send",
    "date_smaller",
    "date_trunc",
    "datetime_pl",
    "datetimetz_pl",
    "dcbrt",
    "default",
    "deferrable",
    "deflate",
    "defrag",
    "degrees",
    "delete_xid_column",
    "delta",
    "delta32k",
    "dense_rank",
    "desc",
    "dexp",
    "diagonal",
    "diameter",
    "disable",
    "dist_cpoly",
    "dist_lb",
    "dist_pb",
    "dist_pc",
    "dist_pl",
    "dist_ppath",
    "dist_ps",
    "dist_sb",
    "dist_sl",
    "distinct",
    "dlog1",
    "dlog10",
    "do",
    "dpow",
    "dround",
    "dsqrt",
    "dtrunc",
    "else",
    "emptyasnull",
    "enable",
    "encode",
    "encrypt     ",
    "encryption",
    "end",
    "eqjoinsel",
    "eqsel",
    "erf",
    "euc_cn_to_mic",
    "euc_jp_to_mic",
    "euc_jp_to_sjis",
    "euc_kr_to_mic",
    "euc_tw_to_big5",
    "euc_tw_to_mic",
    "every",
    "except",
    "exp",
    "explicit",
    "false",
    "for",
    "foreign",
    "freeze",
    "from",
    "full",
    "geometry_analyze",
    "geometry_in",
    "geometry_out",
    "geometry_recv",
    "geometry_send",
    "geometrytype",
    "get_bit",
    "get_byte",
    "getdatabaseencoding",
    "getdate",
    "gistbeginscan",
    "gistbuild",
    "gistbulkdelete",
    "gistcostestimate",
    "gistendscan",
    "gistgettuple",
    "gistinsert",
    "gistmarkpos",
    "gistrescan",
    "gistrestrpos",
    "globaldict256",
    "globaldict64k",
    "grant",
    "group",
    "gzip",
    "hash_aclitem",
    "hashbeginscan",
    "hashbpchar",
    "hashbuild",
    "hashbulkdelete",
    "hashchar",
    "hashcostestimate",
    "hashendscan",
    "hashgettuple",
    "hashinet",
    "hashinsert",
    "hashint2",
    "hashint2vector",
    "hashint4",
    "hashint8",
    "hashmacaddr",
    "hashmarkpos",
    "hashname",
    "hashoid",
    "hashoidvector",
    "hashrescan",
    "hashrestrpos",
    "hashtext",
    "hashvarlena",
    "haversine_distance_km",
    "haversine_distance_km2",
    "having",
    "height",
    "hll_cardinality_16",
    "hll_cardinality_32",
    "host",
    "hostmask",
    "iclikejoinsel",
    "iclikesel",
    "icnlikejoinsel",
    "icnlikesel",
    "icregexeqjoinsel",
    "icregexeqsel",
    "icregexnejoinsel",
    "icregexnesel",
    "identity",
    "ignore",
    "ilike",
    "in",
    "inet",
    "inet_client_addr",
    "inet_client_port",
    "inet_in",
    "inet_out",
    "inet_recv",
    "inet_send",
    "inet_server_addr",
    "inet_server_port",
    "initcap",
    "initially",
    "inner",
    "int2",
    "int24div",
    "int24eq",
    "int24ge",
    "int24gt",
    "int24le",
    "int24lt",
    "int24mi",
    "int24mod",
    "int24mul",
    "int24ne",
    "int24pl",
    "int28div",
    "int28eq",
    "int28ge",
    "int28gt",
    "int28le",
    "int28lt",
    "int28mi",
    "int28mul",
    "int28ne",
    "int28pl",
    "int2_accum",
    "int2_avg_accum",
    "int2_mul_cash",
    "int2_sum",
    "int2abs",
    "int2and",
    "int2div",
    "int2eq",
    "int2ge",
    "int2gt",
    "int2in",
    "int2larger",
    "int2le",
    "int2lt",
    "int2mi",
    "int2mod",
    "int2mul",
    "int2ne",
    "int2not",
    "int2or",
    "int2out",
    "int2pl",
    "int2recv",
    "int2send",
    "int2shl",
    "int2shr",
    "int2smaller",
    "int2um",
    "int2up",
    "int2vectoreq",
    "int2vectorout",
    "int2vectorrecv",
    "int2vectorsend",
    "int2xor",
    "int4",
    "int42div",
    "int42eq",
    "int42ge",
    "int42gt",
    "int42le",
    "int42lt",
    "int42mi",
    "int42mod",
    "int42mul",
    "int42ne",
    "int42pl",
    "int48div",
    "int48eq",
    "int48ge",
    "int48gt",
    "int48le",
    "int48lt",
    "int48mi",
    "int48mul",
    "int48ne",
    "int48pl",
    "int4_accum",
    "int4_avg_accum",
    "int4_mul_cash",
    "int4_sum",
    "int4abs",
    "int4and",
    "int4div",
    "int4eq",
    "int4ge",
    "int4gt",
    "int4in",
    "int4inc",
    "int4larger",
    "int4le",
    "int4lt",
    "int4mi",
    "int4mod",
    "int4mul",
    "int4ne",
    "int4not",
    "int4or",
    "int4out",
    "int4pl",
    "int4recv",
    "int4send",
    "int4shl",
    "int4shr",
    "int4smaller",
    "int4um",
    "int4up",
    "int4xor",
    "int8",
    "int82div",
    "int82eq",
    "int82ge",
    "int82gt",
    "int82le",
    "int82lt",
    "int82mi",
    "int82mul",
    "int82ne",
    "int82pl",
    "int84div",
    "int84eq",
    "int84ge",
    "int84gt",
    "int84le",
    "int84lt",
    "int84mi",
    "int84mul",
    "int84ne",
    "int84pl",
    "int8_accum",
    "int8_avg",
    "int8_avg_accum",
    "int8_pl_timestamp",
    "int8_pl_timestamptz",
    "int8_sum",
    "int8abs",
    "int8and",
    "int8div",
    "int8eq",
    "int8ge",
    "int8gt",
    "int8in",
    "int8inc",
    "int8larger",
    "int8le",
    "int8lt",
    "int8mi",
    "int8mod",
    "int8mul",
    "int8ne",
    "int8not",
    "int8or",
    "int8out",
    "int8pl",
    "int8recv",
    "int8send",
    "int8shl",
    "int8shr",
    "int8smaller",
    "int8um",
    "int8up",
    "int8xor",
    "integer_pl_date",
    "inter_lb",
    "inter_sb",
    "inter_sl",
    "internal_in",
    "internal_out",
    "intersect",
    "interval",
    "interval_accum",
    "interval_avg",
    "interval_cmp",
    "interval_div",
    "interval_eq",
    "interval_ge",
    "interval_gt",
    "interval_hash",
    "interval_in",
    "interval_larger",
    "interval_le",
    "interval_lt",
    "interval_mi",
    "interval_mul",
    "interval_ne",
    "interval_out",
    "interval_pl",
    "interval_pl_date",
    "interval_pl_time",
    "interval_pl_timestamp",
    "interval_pl_timestamptz",
    "interval_pl_timetz",
    "interval_recv",
    "interval_send",
    "interval_smaller",
    "interval_um",
    "intinterval",
    "into",
    "ip_aton",
    "ip_masking",
    "ip_ntoa",
    "is",
    "is_ssl_connection",
    "is_valid_json",
    "is_valid_json_array",
    "isclosed",
    "isnottrue",
    "isnull",
    "iso_to_alt",
    "iso_to_koi8r",
    "iso_to_mic",
    "iso_to_win1251",
    "isopen",
    "isparallel",
    "isperp",
    "istrue",
    "isvertical",
    "join",
    "json_array_length",
    "json_extract_array_element_text",
    "json_extract_path_text",
    "koi8r_to_alt",
    "koi8r_to_iso",
    "koi8r_to_mic",
    "koi8r_to_win1251",
    "lag",
    "language",
    "language_handler_in",
    "language_handler_out",
    "last_day",
    "last_value",
    "latin1_to_mic",
    "latin2_to_mic",
    "latin2_to_win1250",
    "latin3_to_mic",
    "latin4_to_mic",
    "lead",
    "leading",
    "left",
    "len",
    "length",
    "like",
    "like_escape",
    "likejoinsel",
    "likesel",
    "limit",
    "line",
    "line_distance",
    "line_eq",
    "line_in",
    "line_interpt",
    "line_intersect",
    "line_out",
    "line_parallel",
    "line_perp",
    "line_recv",
    "line_send",
    "line_vertical",
    "listagg",
    "listaggdistinct",
    "ln",
    "lo_close",
    "lo_creat",
    "lo_export",
    "lo_import",
    "lo_lseek",
    "lo_open",
    "lo_tell",
    "lo_unlink",
    "localtime",
    "localtimestamp",
    "log",
    "logistic",
    "loread",
    "lower",
    "lpad",
    "lseg",
    "lseg_center",
    "lseg_distance",
    "lseg_eq",
    "lseg_ge",
    "lseg_gt",
    "lseg_in",
    "lseg_interpt",
    "lseg_intersect",
    "lseg_le",
    "lseg_length",
    "lseg_lt",
    "lseg_ne",
    "lseg_out",
    "lseg_parallel",
    "lseg_perp",
    "lseg_recv",
    "lseg_send",
    "lseg_vertical",
    "lun",
    "luns",
    "lzo",
    "lzop",
    "macaddr",
    "macaddr_cmp",
    "macaddr_eq",
    "macaddr_ge",
    "macaddr_gt",
    "macaddr_in",
    "macaddr_le",
    "macaddr_lt",
    "macaddr_ne",
    "macaddr_out",
    "macaddr_recv",
    "macaddr_send",
    "makeaclitem",
    "masklen",
    "max",
    "md5",
    "median",
    "mic_to_alt",
    "mic_to_ascii",
    "mic_to_big5",
    "mic_to_euc_cn",
    "mic_to_euc_jp",
    "mic_to_euc_kr",
    "mic_to_euc_tw",
    "mic_to_iso",
    "mic_to_koi8r",
    "mic_to_latin1",
    "mic_to_latin2",
    "mic_to_latin3",
    "mic_to_latin4",
    "mic_to_sjis",
    "mic_to_win1250",
    "mic_to_win1251",
    "min",
    "minus",
    "mktinterval",
    "mod",
    "months_between",
    "mos_score",
    "mostly13",
    "mostly32",
    "mostly8",
    "mul_d_interval",
    "name",
    "name_pattern_eq",
    "name_pattern_ge",
    "name_pattern_gt",
    "name_pattern_le",
    "name_pattern_lt",
    "name_pattern_ne",
    "nameeq",
    "namege",
    "namegt",
    "nameiclike",
    "nameicnlike",
    "nameicregexeq",
    "nameicregexne",
    "namein",
    "namele",
    "namelike",
    "namelt",
    "namene",
    "namenlike",
    "nameout",
    "namerecv",
    "nameregexeq",
    "nameregexne",
    "namesend",
    "natural",
    "neqjoinsel",
    "neqsel",
    "netmask",
    "network",
    "network_cmp",
    "network_eq",
    "network_ge",
    "network_gt",
    "network_le",
    "network_lt",
    "network_ne",
    "network_sub",
    "network_subeq",
    "network_sup",
    "network_supeq",
    "new",
    "next_day",
    "nextval",
    "nlikejoinsel",
    "nlikesel",
    "node_num",
    "nonnullvalue",
    "not",
    "notlike",
    "notnull",
    "now",
    "npoints",
    "nth_value",
    "ntile",
    "null",
    "nulls",
    "nullvalue",
    "nvl2",
    "octet_length",
    "off",
    "offline",
    "offset",
    "oid",
    "oideq",
    "oidge",
    "oidgt",
    "oidin",
    "oidlarger",
    "oidle",
    "oidlt",
    "oidne",
    "oidout",
    "oidrecv",
    "oidsend",
    "oidsmaller",
    "oidvectoreq",
    "oidvectorge",
    "oidvectorgt",
    "oidvectorle",
    "oidvectorlt",
    "oidvectorne",
    "oidvectorout",
    "oidvectorrecv",
    "oidvectorsend",
    "oidvectortypes",
    "old",
    "on",
    "on_pb",
    "on_pl",
    "on_ppath",
    "on_ps",
    "on_sb",
    "on_sl",
    "only",
    "opaque_in",
    "opaque_out",
    "open",
    "or",
    "order",
    "outer",
    "overlaps",
    "parallel",
    "partition",
    "path",
    "path_add",
    "path_add_pt",
    "path_center",
    "path_contain_pt",
    "path_distance",
    "path_div_pt",
    "path_in",
    "path_inter",
    "path_length",
    "path_mul_pt",
    "path_n_eq",
    "path_n_ge",
    "path_n_gt",
    "path_n_le",
    "path_n_lt",
    "path_npoints",
    "path_out",
    "path_recv",
    "path_send",
    "path_sub_pt",
    "pclose",
    "percent",
    "percent_rank",
    "percentile_cont",
    "percentile_disc",
    "permissions",
    "pg_backend_pid",
    "pg_cancel_backend",
    "pg_char_to_encoding",
    "pg_client_encoding",
    "pg_conversion_is_visible",
    "pg_decode",
    "pg_encoding_to_char",
    "pg_get_cols",
    "pg_get_constraintdef",
    "pg_get_current_groups",
    "pg_get_expr",
    "pg_get_external_columns",
    "pg_get_external_tables",
    "pg_get_indexdef",
    "pg_get_late_binding_view_cols",
    "pg_get_ruledef",
    "pg_get_userbyid",
    "pg_get_viewdef",
    "pg_last_copy_count",
    "pg_last_copy_id",
    "pg_last_query_id",
    "pg_last_unload_count",
    "pg_last_unload_id",
    "pg_lock_status",
    "pg_opclass_is_visible",
    "pg_operator_is_visible",
    "pg_postmaster_start_time",
    "pg_show_all_settings",
    "pg_sleep",
    "pg_start_backup",
    "pg_stat_get_backend_activity",
    "pg_stat_get_backend_activity_start",
    "pg_stat_get_backend_dbid",
    "pg_stat_get_backend_idset",
    "pg_stat_get_backend_pid",
    "pg_stat_get_blocks_hit",
    "pg_stat_get_db_blocks_hit",
    "pg_stat_get_db_numbackends",
    "pg_stat_get_db_xact_commit",
    "pg_stat_get_db_xact_rollback",
    "pg_stat_get_numscans",
    "pg_stat_get_tuples_deleted",
    "pg_stat_get_tuples_inserted",
    "pg_stat_get_tuples_returned",
    "pg_stat_get_tuples_updated",
    "pg_stat_reset",
    "pg_stop_backup",
    "pg_table_is_visible",
    "pg_tablespace_databases",
    "pg_terminate_backend",
    "pg_timezone_abbrevs",
    "pg_timezone_names",
    "pg_type_is_visible",
    "pgdate_part",
    "pi",
    "placing",
    "plpython_call_handler",
    "plpython_compiler",
    "point",
    "point_above",
    "point_add",
    "point_below",
    "point_distance",
    "point_div",
    "point_eq",
    "point_in",
    "point_mul",
    "point_ne",
    "point_out",
    "point_recv",
    "point_send",
    "point_sub",
    "point_vert",
    "poly_center",
    "poly_contain",
    "poly_contain_pt",
    "poly_contained",
    "poly_distance",
    "poly_in",
    "poly_npoints",
    "poly_out",
    "poly_overlap",
    "poly_recv",
    "poly_same",
    "poly_send",
    "polygon",
    "popen",
    "position",
    "positionjoinsel",
    "positionsel",
    "pow",
    "power",
    "primary",
    "pt_contained_circle",
    "pt_contained_poly",
    "qsummary",
    "quote_ident",
    "quote_literal",
    "radians",
    "radius",
    "random",
    "rank",
    "rankx",
    "ratio_to_report",
    "raw",
    "readratio",
    "record_in",
    "record_out",
    "record_recv",
    "record_send",
    "recover",
    "references",
    "regclassin",
    "regclassout",
    "regclassrecv",
    "regclasssend",
    "regexeqjoinsel",
    "regexeqsel",
    "regexnejoinsel",
    "regexnesel",
    "regexp_count",
    "regexp_instr",
    "regexp_replace",
    "regexp_substr",
    "regoperatorout",
    "regoperatorrecv",
    "regoperatorsend",
    "regoperout",
    "regoperrecv",
    "regopersend",
    "regprocedurein",
    "regprocedureout",
    "regprocedurerecv",
    "regproceduresend",
    "regprocin",
    "regprocout",
    "regprocrecv",
    "regprocsend",
    "regtypein",
    "regtypeout",
    "regtyperecv",
    "regtypesend",
    "rejectlog",
    "reltime",
    "reltimeeq",
    "reltimege",
    "reltimegt",
    "reltimein",
    "reltimele",
    "reltimelt",
    "reltimene",
    "reltimeout",
    "reltimerecv",
    "reltimesend",
    "relu",
    "remaining_vacrows",
    "repeat",
    "replace",
    "replicate",
    "resort",
    "respect",
    "restore",
    "return_approx_percentile",
    "reverse",
    "right",
    "round",
    "row_number",
    "rownumber",
    "rpad",
    "rt_box_inter",
    "rt_box_size",
    "rt_box_union",
    "rt_poly_inter",
    "rt_poly_size",
    "rt_poly_union",
    "rtbeginscan",
    "rtbuild",
    "rtbulkdelete",
    "rtcostestimate",
    "rtendscan",
    "rtgettuple",
    "rtinsert",
    "rtmarkpos",
    "rtrescan",
    "rtrestrpos",
    "scalargtjoinsel",
    "scalargtsel",
    "scalarltjoinsel",
    "scalarltsel",
    "select",
    "session_user",
    "set_bit",
    "set_byte",
    "set_masklen",
    "setval",
    "sig_bucket_3g",
    "sig_bucket_lte",
    "sig_score_3g",
    "sig_score_lte",
    "sigmoid",
    "sign",
    "similar",
    "similar_escape",
    "sin",
    "sjis_to_euc_jp",
    "sjis_to_mic",
    "slice_num",
    "slope",
    "smgreq",
    "smgrne",
    "smgrout",
    "snapshot ",
    "some",
    "sort_avail_mem",
    "sortbytes",
    "split_part",
    "sqrt",
    "ssl_version",
    "st_asbinary",
    "st_asewkb",
    "st_asewkt",
    "st_astext",
    "st_azimuth",
    "st_contains",
    "st_dimension",
    "st_disjoint",
    "st_distance",
    "st_dwithin",
    "st_equals",
    "st_geometrytype",
    "st_intersects",
    "st_isclosed",
    "st_iscollection",
    "st_isempty",
    "st_makeline",
    "st_makepoint",
    "st_makepolygon",
    "st_npoints",
    "st_numpoints",
    "st_point",
    "st_polygon",
    "st_within",
    "st_x",
    "st_xmax",
    "st_xmin",
    "st_y",
    "st_ymax",
    "st_ymin",
    "stddev",
    "stddev_pop",
    "strpos",
    "strtol",
    "substr",
    "sum",
    "sysdate",
    "system",
    "table",
    "tablebytes",
    "tag",
    "tan",
    "tdes",
    "text",
    "text255",
    "text32k",
    "text_ge",
    "text_gt",
    "text_larger",
    "text_le",
    "text_lt",
    "text_pattern_eq",
    "text_pattern_ge",
    "text_pattern_gt",
    "text_pattern_le",
    "text_pattern_lt",
    "text_pattern_ne",
    "text_smaller",
    "textcat",
    "texteq",
    "texticlike",
    "texticnlike",
    "texticregexeq",
    "texticregexne",
    "textin",
    "textlen",
    "textlike",
    "textne",
    "textnlike",
    "textout",
    "textrecv",
    "textregexeq",
    "textregexne",
    "textsend",
    "then",
    "tideq",
    "tidin",
    "tidout",
    "tidrecv",
    "tidsend",
    "time",
    "time_cmp",
    "time_eq",
    "time_ge",
    "time_gt",
    "time_in",
    "time_larger",
    "time_le",
    "time_lt",
    "time_mi_interval",
    "time_mi_time",
    "time_ne",
    "time_out",
    "time_pl_interval",
    "time_recv",
    "time_send",
    "time_smaller",
    "timedate_pl",
    "timemi",
    "timenow",
    "timepl",
    "timestamp",
    "timestamp_cmp",
    "timestamp_cmp_date",
    "timestamp_cmp_timestamptz",
    "timestamp_eq",
    "timestamp_eq_date",
    "timestamp_eq_timestamptz",
    "timestamp_ge",
    "timestamp_ge_date",
    "timestamp_ge_timestamptz",
    "timestamp_gt",
    "timestamp_gt_date",
    "timestamp_gt_timestamptz",
    "timestamp_in",
    "timestamp_larger",
    "timestamp_le",
    "timestamp_le_date",
    "timestamp_le_timestamptz",
    "timestamp_lt",
    "timestamp_lt_date",
    "timestamp_lt_timestamptz",
    "timestamp_mi",
    "timestamp_mi_int8",
    "timestamp_mi_interval",
    "timestamp_ne",
    "timestamp_ne_date",
    "timestamp_ne_timestamptz",
    "timestamp_out",
    "timestamp_pl_int8",
    "timestamp_pl_interval",
    "timestamp_recv",
    "timestamp_send",
    "timestamp_smaller",
    "timestamptz",
    "timestamptz_cmp",
    "timestamptz_cmp_date",
    "timestamptz_cmp_timestamp",
    "timestamptz_eq",
    "timestamptz_eq_date",
    "timestamptz_eq_timestamp",
    "timestamptz_ge",
    "timestamptz_ge_date",
    "timestamptz_ge_timestamp",
    "timestamptz_gt",
    "timestamptz_gt_date",
    "timestamptz_gt_timestamp",
    "timestamptz_in",
    "timestamptz_larger",
    "timestamptz_le",
    "timestamptz_le_date",
    "timestamptz_le_timestamp",
    "timestamptz_lt",
    "timestamptz_lt_date",
    "timestamptz_lt_timestamp",
    "timestamptz_mi",
    "timestamptz_mi_int8",
    "timestamptz_mi_interval",
    "timestamptz_ne",
    "timestamptz_ne_date",
    "timestamptz_ne_timestamp",
    "timestamptz_out",
    "timestamptz_pl_int8",
    "timestamptz_pl_interval",
    "timestamptz_recv",
    "timestamptz_send",
    "timestamptz_smaller",
    "timetz",
    "timetz_cmp",
    "timetz_eq",
    "timetz_ge",
    "timetz_gt",
    "timetz_hash",
    "timetz_in",
    "timetz_larger",
    "timetz_le",
    "timetz_lt",
    "timetz_mi_interval",
    "timetz_ne",
    "timetz_out",
    "timetz_pl_interval",
    "timetz_recv",
    "timetz_send",
    "timetz_smaller",
    "timetzdate_pl",
    "timezone",
    "tinterval",
    "tintervalct",
    "tintervalend",
    "tintervaleq",
    "tintervalge",
    "tintervalgt",
    "tintervalin",
    "tintervalle",
    "tintervalleneq",
    "tintervallenge",
    "tintervallengt",
    "tintervallenle",
    "tintervallenlt",
    "tintervallenne",
    "tintervallt",
    "tintervalne",
    "tintervalout",
    "tintervalov",
    "tintervalrecv",
    "tintervalrel",
    "tintervalsame",
    "tintervalsend",
    "tintervalstart",
    "to",
    "to_ascii",
    "to_char",
    "to_date",
    "to_hex",
    "to_number",
    "to_timestamp",
    "tobpchar",
    "top",
    "total_num_deleted_vacrows",
    "total_num_deleted_vacrows_reclaimed",
    "trailing",
    "translate",
    "true",
    "trunc",
    "truncatecolumns",
    "union",
    "unique",
    "unknownin",
    "unknownout",
    "unknownrecv",
    "unknownsend",
    "upper",
    "user",
    "using",
    "vacstate_endrow",
    "vacstate_last_inserted_row",
    "var_pop",
    "varbit",
    "varbit_in",
    "varbit_out",
    "varbit_recv",
    "varbit_send",
    "varbitcmp",
    "varbiteq",
    "varbitge",
    "varbitgt",
    "varbitle",
    "varbitlt",
    "varbitne",
    "varchar",
    "varcharout",
    "varcharrecv",
    "varcharsend",
    "verbose",
    "version",
    "void_in",
    "void_out",
    "wallet",
    "when",
    "where",
    "width",
    "width_bucket",
    "win1250_to_latin2",
    "win1250_to_mic",
    "win1250_to_utf",
    "win1251_to_alt",
    "win1251_to_iso",
    "win1251_to_koi8r",
    "win1251_to_mic",
    "win1256_to_utf",
    "win874_to_utf",
    "with",
    "without",
    "xideq",
    "xideqint4",
    "xidin",
    "xidout",
    "xidrecv",
    "xidsend",
}



# -*- coding: utf-8 -*-

from typing import Any, Dict, Generator, List, Optional, Tuple, Union


def pk(parent: Optional[Union[Tuple[str], List[str]]], key: str) -> Tuple[str]:
    """Construct chained parent key as list of strings."""

    if parent:
        return tuple((*parent, key))
    return tuple((key,))


def terminal(v: Any) -> bool:
    """Detect terminal key ending in type other than non-empty dict or list."""

    if isinstance(v, (dict, list)):
        if v:
            return False
    return True


def extract_terminal_keys(
    d: Dict, parent: Optional[List[str]] = None
) -> Generator[Tuple[Tuple[str], Dict], None, None]:
    """Get parent keys that reference a value or no other keys."""

    if isinstance(d, dict):
        for k, v in d.items():
            parent_key = pk(parent, k)
            if terminal(v):
                if isinstance(v, list):
                    yield pk(parent_key, "[array]"), None
                else:
                    yield parent_key, v
            else:
                yield from extract_terminal_keys(v, parent_key)
    elif isinstance(d, list):
        if d:
            for item in d:
                parent_key = pk(parent, "[array]")
                yield from extract_terminal_keys(item, parent_key)
        else:
            parent_key = pk(parent, "[array]")
            yield parent_key, None
    else:
        yield parent, d


# Create Dict --------------------------------------------------------------------------


def split_key(parent_key: List[str]) -> List[List[str]]:
    """Split group key into parts."""

    key_parts = []
    while parent_key:
        key_parts.append(parent_key)
        parent_key = parent_key[:-1]
    return key_parts[::-1]


def construct_branch(
    d: Dict, loc: Dict, group_key: Tuple[str], is_dict: bool = False, key_val=None
):
    """Construct dict branch from terminal parent key."""

    key_parts = split_key(group_key)
    array_keys = [k for k in key_parts if k[-1] == "[array]"]
    array_parents = [k[:-1] for k in array_keys]

    for key_part in key_parts:
        if key_part in loc:
            continue

        # detect top level key first
        if len(key_part) == 1:
            key = key_part[0]

            if key_part == group_key:
                val = {} if is_dict else key_val
            elif key_part in array_parents:
                val = []
            else:
                val = {}

            d[key] = val
            loc[key_part] = d[key]
        else:
            par_key, key = key_part[:-1], key_part[-1]
            ref = loc[par_key]

            if key_part in array_parents:
                # child key_part ends in dict ref or [array]
                # array parents create list + point to list
                # implies non-terminal key
                val = []
                if par_key in array_parents:
                    # reference is list
                    # some_key.[array].[array]
                    if key_part in array_keys:
                        # some_key.[array].[array]
                        ref.append(val)
                        loc[key_part] = loc[par_key][0]
                    else:
                        # some_key.[array]
                        sub_d = {key: val}
                        if ref:
                            ref[0].update(sub_d)
                        else:
                            ref.append(sub_d)
                        loc[key_part] = ref[0][key]
                elif par_key in array_keys:
                    if key_part in array_keys:
                        loc[key_part] = ref
                    else:
                        sub_d = {key: val}
                        if ref:
                            ref[0].update(sub_d)
                        else:
                            ref.append(sub_d)
                        loc[key_part] = ref[0][key]
                else:
                    # some_key.[array] == {k: []}
                    sub_d = {key: val}
                    ref.update(sub_d)
                    loc[key_part] = ref[key]

            elif key_part in array_keys:
                # append to parent ref
                if key_part == group_key:
                    # terminal
                    if key_val is not None:
                        ref.append(key_val)
                else:
                    loc[key_part] = ref
            elif key_part == group_key:
                # assign key_val
                sub_d = {key: key_val}
                if par_key in array_keys:
                    if ref:
                        ref[0].update(sub_d)
                    else:
                        ref.append(sub_d)
                else:
                    ref.update(sub_d)
            else:
                val = {}
                sub_d = {key: val}
                if par_key in array_keys:
                    if ref:
                        ref[0].update(sub_d)
                    else:
                        ref.append(sub_d)
                    ref = ref[0]
                else:
                    ref.update(sub_d)

                loc[key_part] = ref[key]

    return d, loc


def reconstruct_dict(key_store: Dict) -> Dict:
    """Construct dict from key store."""

    d, loc = {}, {}

    for group_key, field in sorted(key_store.items(), key=lambda t: t[0]):

        val = field.max_value
        is_dict = isinstance(val, dict)
        if val == []:
            val = None

        construct_branch(d, loc, group_key, is_dict=is_dict, key_val=val)

    return d, loc



# -*- coding: utf-8 -*-

import logging

from collections import defaultdict
from itertools import chain
from typing import Dict

logger = logging.getLogger(__name__)


def add_hist_dicts(d1: Dict, d2: Dict) -> Dict:
    d = defaultdict(int)
    for k, v in chain(d1.items(), d2.items()):
        d[k] += v
    return d


class Field:
    """Tracks `max` value and data type(s)."""

    numeric_types = {"int", "float"}

    def __init__(self, parent_key: str, value=None, *, str_numeric_override=False):
        self.parent_key = parent_key
        self.str_numeric_override = str_numeric_override
        self.num_na = 0
        self.dtype_max = {}
        self.hist = defaultdict(int)
        self.add(value)

    def __add__(self, other):
        str_numeric_override = any(
            (self.str_numeric_override, other.str_numeric_override)
        )
        field = Field(self.parent_key, str_numeric_override=str_numeric_override)
        field.num_na = self.num_na + other.num_na

        for dtype, value in chain(self.dtype_max.items(), other.dtype_max.items()):
            field._update_dtype_max(dtype, value)

        hist = add_hist_dicts(self.hist, other.hist)
        field.hist.update(hist)
        return field

    def push_warnings(self):
        """Detect and log mixed dtypes."""

        if len(self.hist.keys()) > 1:
            if isinstance(self.parent_key, tuple):
                ref_par_key = ".".join(self.parent_key)
            else:
                ref_par_key = self.parent_key

            logger.warning(
                f"[{ref_par_key}] dtypes detected {', '.join(sorted(self.hist.keys()))}"
            )

    @property
    def dtype(self):
        """Get data type conforming to spectrum requirements or with highest count.

        If `str_numeric_override` is enabled and any strings have been seen, returned
        dtype is forced as str.

        If int and float have been seen, dtype defaults to float.
        """

        dtype = None
        if self.hist:

            if len(self.hist.keys()) == 1:
                dtype = list(self.hist.keys())[0]
            else:
                numeric_intersect = self.numeric_types & self.hist.keys()

                if numeric_intersect:
                    if self.str_numeric_override and "str" in self.hist:
                        dtype = "str"
                    elif "float" in numeric_intersect:
                        dtype = "float"

            if not dtype:
                dtype, _ = max(self.hist.items(), key=lambda t: t[1])

        return dtype

    def _get_max_comparable(self, prev_value, value, key_func):
        """Get max value for inputs which can be compared."""

        max_value = None
        if prev_value is None:
            max_value = value
        else:
            max_value = max(value, prev_value, key=key_func)
        return max_value

    def _compare_numeric(self, prev_value, value):
        return self._get_max_comparable(prev_value, value, abs)

    def _compare_str(self, prev_value, value):
        return self._get_max_comparable(prev_value, value, len)

    def _compare_other_types(self, prev_value, value):
        return value

    @property
    def max_value(self):
        """Get max value for current dtype."""

        _dtype = self.dtype
        is_numeric = _dtype in self.numeric_types

        if is_numeric and self.numeric_types.issubset(self.hist.keys()):
            val = self._compare_numeric(self.dtype_max["float"], self.dtype_max["int"])
            return float(val)

        return self.dtype_max.get(_dtype)

    def _update_dtype_max(self, incoming_dtype, value):
        """Detect dtype change and store diffs."""

        if incoming_dtype in self.dtype_max:
            prev_value = self.dtype_max[incoming_dtype]

            comp_func = None
            if incoming_dtype == "str":
                comp_func = self._compare_str
            elif incoming_dtype in self.numeric_types:
                comp_func = self._compare_numeric
            else:
                comp_func = self._compare_other_types

            self.dtype_max[incoming_dtype] = comp_func(prev_value, value)

        else:
            self.dtype_max[incoming_dtype] = value

    def add(self, value):
        """Add value to field and track dtype."""

        if value is not None:
            incoming_dtype = type(value).__name__
            self._update_dtype_max(incoming_dtype, value)
            self.hist[incoming_dtype] += 1
        else:
            self.num_na += 1



# -*- coding: utf-8 -*-

import codecs
import os.path
import re
import setuptools


def read(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


requirements = ["black>=19.10b0", "pendulum>=1.4.4"]


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spectron",
    version=find_version("spectron", "__init__.py"),
    author="Jeremy Jacobs",
    author_email="pub@j4c0bs.net",
    description="AWS Redshift Spectrum utilities.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/j4c0bs/spectron",
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    install_requires=requirements,
    tests_require=["pytest", "pytest-datadir"],
    extras_require={"json": "ujson==1.35"},
    include_package_data=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={"console_scripts": ["spectron=spectron.cli:create_spectrum_schema"]},
)



# -*- coding: utf-8 -*-

import logging

from collections import defaultdict
from itertools import chain
from typing import AbstractSet, Dict, Generator, List, Optional, Tuple

from . import data_types
from .Field import Field, add_hist_dicts
from .merge import construct_branch, extract_terminal_keys

logger = logging.getLogger(__name__)

# Mixed array group utils --------------------------------------------------------------


def get_array_parents(field_key: Tuple[str]) -> Generator[Tuple[str], None, None]:
    """Get all array parent keys."""

    ix = 0
    for _ in range(field_key.count("[array]")):
        ix = field_key.index("[array]", ix)
        yield field_key[:ix]
        ix += 1


def loc_siblings(
    parent_key: Tuple[str], keys: List[Tuple[str]], is_array: Optional[bool] = None
) -> List[Tuple[str]]:
    """Locate Field keys which share a parent key."""

    if is_array is not None:
        if is_array:
            comp = lambda k: k == "[array]"
        else:
            comp = lambda k: k != "[array]"
    else:
        comp = lambda k: True

    res = []
    num_parents = len(parent_key)
    for k in keys:
        if len(k) > num_parents:
            if k[:num_parents] == parent_key and comp(k[num_parents]):
                res.append(k)
    return res


def is_parent(parent_key: Tuple[str], key: Tuple[str]) -> bool:
    """Determines if keys have parent:child relationship.

    Example using dot notation:
        is_parent(a.b, a.b.c) == True
        is_parent(a.b, a.b.c.d) == True
        is_parent(a.b, a.x.y) == False

    """

    return len(key) > len(parent_key) and key[: len(parent_key)] == parent_key


# --------------------------------------------------------------------------------------


class MaxDict:
    """Collect and store field, `max` value per key branch.

    When generating a dict, mixed data types and mixed array keys are resolved.

    Data types:
        For fields which have seen string and numeric, priority is determined via:
            str > float > int

    Array keys:
        If a parent key contains both array and non-array child keys, the number of
        values seen for all downstream branches are summed and compared. The key group
        with the highest count will override the other. If counts are equal, array keys
        are prioritized.

        Example using dot notation:

            terminal keys:
                - a.b.[array].d.e = 10
                - a.b.[array].d.f = 10
                - a.b.c.x = 5
                - a.b.c.y = 5

            Count of a.b.[array] = 20, count of a.b.c = 10.
            The output of MaxDict.asdict() will contain a.b.[array] and not a.b.c
    """

    def __init__(self, str_numeric_override: bool = False):
        self.str_numeric_override = str_numeric_override
        self.hist = defaultdict(int)
        self.key_store = {}
        self.override_keys = None

    def __add__(self, other):
        str_numeric_override = any(
            (self.str_numeric_override, other.str_numeric_override)
        )
        md = MaxDict(str_numeric_override=str_numeric_override)
        md.hist.update(add_hist_dicts(self.hist, other.hist))

        for key in self.key_store.keys() & other.key_store.keys():
            md.key_store[key] = self.key_store[key] + other.key_store[key]

        for key in self.key_store.keys() ^ other.key_store.keys():
            if key in self.key_store:
                field = self.key_store[key]
            else:
                field = other.key_store[key]

            md.key_store[key] = field + Field(key)  # coerce new instance

        return md

    def add(self, key: str, value):
        self.hist[key] += 1
        if key in self.key_store:
            self.key_store[key].add(value)
        else:
            self.key_store[key] = Field(
                key, value, str_numeric_override=self.str_numeric_override
            )

    def load_dict(self, d: Dict):
        for key, value in extract_terminal_keys(d):
            self.add(key, value)

    def batch_load_dicts(self, items: List[Dict]):
        for d in items:
            self.load_dict(d)

    def fields(self):
        yield from self.key_store.values()

    def fields_seen(self) -> Tuple[int, int]:
        tot = 0
        tot_na = 0
        for field in self.fields():
            tot += sum(field.hist.values())
            tot_na += field.num_na
        return tot, tot_na

    def has_dtype_changes(self) -> List[Field]:
        loc = []
        for field in self.fields():
            if field.dtype_change:
                loc.append(field)
        return loc

    def asdict(self, astype: bool = False) -> Dict:
        """Resolved keys, [max | type] vals as dict."""

        self.load_override_keys()

        d, loc = {}, {}
        for group_key, field in sorted(self.key_store.items(), key=lambda t: t[0]):

            if group_key in self.override_keys:
                continue

            is_dict = False
            val = field.max_value

            if val is None:
                logger.warning(f"Ignoring key with None value: {'.'.join(group_key)}")
                continue

            if isinstance(val, dict):
                is_dict = True
            elif isinstance(val, list):
                if val and val[0] is not None:
                    val = val.pop(0)
                else:
                    val = []
            elif astype:
                val = data_types.set_dtype(val)

            construct_branch(d, loc, group_key, is_dict=is_dict, key_val=val)

        return d

    # detect mixed terminal, array - dict keys -----------------------------------------

    def sum_key_groups(self, parent_key: Tuple[str]) -> Dict:
        """Sums Field value counts in array, non-array branch groups."""

        key_groups = {
            "array": {"keys": [], "total": 0},
            "non_array": {"keys": [], "total": 0},
        }

        par_ix = len(parent_key)
        for key, field in self.key_store.items():
            if is_parent(parent_key, key):
                count = sum(field.hist.values())

                if key[par_ix] == "[array]":
                    group_ref = key_groups["array"]
                else:
                    group_ref = key_groups["non_array"]

                group_ref["total"] += count
                group_ref["keys"].append(key)

        return key_groups

    def detect_mixed_array_parents(
        self, keys: List[Tuple[str]]
    ) -> AbstractSet[Tuple[str]]:
        """Detect parent keys which have array and non-array children.

        Example using dot notation:

            terminal keys:
                - a.b.[array].d
                - a.b.c.d
            mixed array parent:
                - a.b
        """

        array_keys = [k for k in keys if "[array]" in k]
        array_parents = sorted(
            set(chain(*map(get_array_parents, array_keys))), key=lambda t: len(t)
        )

        mixed_array_parents = set()
        for parent_key in array_parents:
            non_array_siblings = loc_siblings(parent_key, keys, is_array=False)
            if non_array_siblings:
                mixed_array_parents.add(parent_key)

        return mixed_array_parents

    def detect_mixed_terminal_keys(self, keys: List[Tuple[str]]) -> List[Tuple[str]]:
        """Detect terminal keys which are also parent keys.

        Example using dot notation:
            - a.b (added to override_keysride)
            - a.b.c
        """

        max_num_keys = max(len(k) for k in keys)
        mixed_terminal_keys = []

        for parent_key in (k for k in keys if len(k) < max_num_keys):
            parent_key_len = len(parent_key)
            for child_key in (k for k in keys if len(k) > parent_key_len):
                if is_parent(parent_key, child_key):
                    mixed_terminal_keys.append(parent_key)
                    break

        return sorted(mixed_terminal_keys, key=lambda t: len(t))

    def load_override_keys(self):
        """Inspect mixed array groups and select group with highest count.

        If counts are equal, array keys take precedence and non-array keys are ignored
        when constructing dict.
        """

        self.override_keys = []
        keys = sorted(self.key_store.keys(), key=lambda t: len(t))
        mixed_terminal_keys = self.detect_mixed_terminal_keys(keys)

        if mixed_terminal_keys:
            self.override_keys.extend(mixed_terminal_keys)

            # filter terminal keys before detecting mixed array keys
            keys = [k for k in keys if k not in mixed_terminal_keys]

            for terminal_key in mixed_terminal_keys:
                logger.warning(f"Terminal key detected: {'.'.join(terminal_key)}")

        mixed_array_parents = self.detect_mixed_array_parents(keys)
        for parent_key in sorted(mixed_array_parents, key=lambda t: len(t)):
            key_groups = self.sum_key_groups(parent_key)

            pk_ref = ".".join(parent_key)
            num_array = key_groups["array"]["total"]
            num_non_array = key_groups["non_array"]["total"]

            if num_array >= num_non_array:
                self.override_keys.extend(key_groups["non_array"]["keys"])

                logger.warning(
                    f"{pk_ref}: Array keys override non-arrays: {num_array:,} >= {num_non_array:,}"
                )
            else:
                self.override_keys.extend(key_groups["array"]["keys"])

                logger.warning(
                    f"{pk_ref}: Non-Array keys override arrays: {num_non_array:,} >= {num_array:,}"
                )

        self.override_keys = set(self.override_keys)



# -*- coding: utf-8 -*-

import logging
import sys

from functools import partial
from itertools import takewhile

try:
    import ujson as json
except ImportError:
    import json

from . import data_types
from . import reserved
from . import write_ddl


logger = logging.getLogger(__name__)

# --------------------------------------------------------------------------------------


def _count_indent(line):
    for _ in takewhile(lambda c: c == " ", line):
        yield 1


def strip_top_level_seps(s):
    lines = []
    for line in s.split("\n"):
        if sum(_count_indent(line)) == 4:
            line = line.replace(":", " ")
        lines.append(line)
    return "\n".join(lines)


def _conform_syntax(d):
    """Replace Python syntax to match Spectrum DDL syntax."""

    s = json.dumps(d, indent=4).strip()

    # remove outermost brackets
    s = s[1:][:-1]
    s = s.rstrip()

    # replace dict, lists with schema dtypes
    s = s.replace("{", "struct<").replace("}", ">")
    s = s.replace("[", "array<").replace("]", ">")

    # drop colons in top level fields
    s = strip_top_level_seps(s)

    # add space after colon
    s = s.replace(":", ": ")

    # drop quotes, replace back-ticks
    s = s.replace('"', "").replace("`", '"')
    return s


# Keys ---------------------------------------------------------------------------------


def _as_parent_key(parent, key):
    """Construct parent key."""

    if parent:
        return f"{parent}.{key}"
    return key


def validate_identifier(key: str):
    """Confirm key is valid identifier.
        - between 1 and 127 bytes in length
        - does not contain quotation marks
        - starts with alphabetic or underscore character
            * does not raise error, fixed in mapping
    """

    key = key.strip()

    if not key:
        raise ValueError("Column name is empty string...")

    if len(key) > 127:
        raise ValueError("Column name exceeds 127 characters...")

    if "'" in key or '"' in key:
        raise ValueError("Column name contains quotation marks...")

    return key[0].isalpha() or key.startswith("_")


def detect_reserved_key(key, parent):
    if key.strip("`").lower() in reserved.keywords:
        logger.info(f"Reserved keyword detected: {parent}.{key}")


def detect_hyphens(key: str, convert_hyphens: bool):
    """Detect hyphens and enclose in quotes if located."""

    if "-" in key:
        if convert_hyphens:
            key = key.replace("-", "_")
        else:
            key = f"`{key}`"
    return key


def conform_key(
    key: str,
    mapping: dict,
    case_map: bool,
    convert_hyphens: bool,
    case_insensitive: bool,
):
    """Conform key to user options and standard identifier rules."""

    if case_insensitive and case_map:
        raise ValueError("case_insensitive and case_map both True...")

    mapped_key = None

    use_key_map = mapping and key in mapping
    is_reserved = key.lower() in reserved.keywords

    if use_key_map:
        mapped_key = mapping[key]

    if is_reserved and not mapped_key:
        if case_insensitive:
            key = key.lower()
        key = f"`{key}`"
        return key, mapped_key

    # apply case fold to source column and user defined mapped key
    if case_map:
        if mapped_key:
            if any(c.isupper() for c in mapped_key):
                mapped_key = mapped_key.lower()
        elif any(c.isupper() for c in key):
            mapped_key = key.lower()
    elif case_insensitive:
        key = key.lower()
        if mapped_key:
            mapped_key = mapped_key.lower()

    # check for invalid keys and mutate key used in DDL
    if mapped_key:
        if not validate_identifier(mapped_key):
            mapped_key = f"_{mapped_key}"
    elif not validate_identifier(key):
        mapped_key = f"_{key}"

    # override user map and enclose hyphenated column name in quotes
    if mapped_key:
        mapped_key = detect_hyphens(mapped_key, convert_hyphens)
    else:
        proc_key = detect_hyphens(key, convert_hyphens)
        if proc_key.strip("`") != key:
            mapped_key = proc_key
        else:
            key = proc_key

    return key, mapped_key


def process_keys(
    parent,
    keys,
    mapping,
    ignore_fields,
    case_map,
    convert_hyphens,
    case_insensitive,
    **kwargs,
):
    """Apply key transforms and detect conflicts."""

    conform = partial(
        conform_key,
        mapping=mapping,
        case_map=case_map,
        convert_hyphens=convert_hyphens,
        case_insensitive=case_insensitive,
    )

    keys = [k for k in keys if not (ignore_fields and k in ignore_fields)]
    proc_keys = {key: conform(key) for key in keys}

    new_keys = [k[1] if k[1] is not None else k[0] for k in proc_keys.values()]

    # detect key conflict
    if len(set(new_keys)) < len(keys):
        conflicts = sorted(set(k for k in new_keys if new_keys.count(k) > 1))
        raise ValueError(f"Key conflicts in {parent}: {', '.join(conflicts)}")

    return proc_keys


# --------------------------------------------------------------------------------------


def count_members(d):
    total = 0
    if d is not None:
        if isinstance(d, dict):
            total += len(d.keys())
            for v in d.values():
                total += count_members(v)
        elif isinstance(d, list):
            total += len(d)
            for item in d:
                if isinstance(item, dict):
                    total += count_members(item)
        else:
            total += 1
    return total


def validate_array(array, parent, ignore_nested_arrarys):
    """Confirm array has single data type, no empty or nested arrays."""

    if not array:
        logger.warning(f"[{parent}] Ignoring empty array")
        return False

    # confirm single dtype
    _num_dtypes = len(data_types.type_set(array))
    if _num_dtypes > 1:
        logger.warning(
            f"[{parent}] Ignoring array with multiple dtypes ({_num_dtypes})"
        )
        return False

    # check for nested arrays
    if any(isinstance(item, list) for item in array):
        if ignore_nested_arrarys:
            logger.warning(f"[{parent}] Ignoring nested arrays ({len(array)})")
            return False
        else:
            msg = f"Nested arrays detected ({len(array)}) in {parent}..."
            raise ValueError(msg)

    return True


def define_types(
    d,
    mapping=None,
    type_map=None,
    ignore_fields=None,
    infer_date=False,
    convert_hyphens=False,
    case_map=False,
    case_insensitive=False,
    ignore_nested_arrarys=True,
):
    """Replace values with data types and maintain data structure."""

    key_map = {}  # dict of confirmed keys to include in serde mapping

    kwargs = {
        "mapping": mapping,
        "type_map": type_map,
        "ignore_fields": ignore_fields,
        "infer_date": infer_date,
        "convert_hyphens": convert_hyphens,
        "case_map": case_map,
        "case_insensitive": case_insensitive,
        "ignore_nested_arrarys": ignore_nested_arrarys,
    }

    def parse_types(d, parent=None):
        """Crawl and assign data types."""

        nonlocal key_map

        if isinstance(d, list):
            as_types = []
            parent_key = _as_parent_key(parent, "array")

            if not validate_array(d, parent_key, ignore_nested_arrarys):
                return None

            if any(isinstance(item, dict) for item in d):
                if len(d) == 1:
                    single_dict = d[0]
                else:
                    single_dict = sorted(d, key=count_members, reverse=True)[0]

                as_types.append(parse_types(single_dict, parent=parent_key))

            else:
                for item in d:
                    if isinstance(item, list):
                        continue
                    as_types.append(parse_types(item, parent=parent_key))
                as_types = sorted(set(as_types))

        elif isinstance(d, dict):
            as_types = {}

            proc_keys = process_keys(parent, d.keys(), **kwargs)

            for key, val in d.items():
                if ignore_fields and key in ignore_fields:
                    continue

                dtype = None
                parent_key = _as_parent_key(parent, key)
                ref_key, mapped_key = proc_keys[key]

                # set user defined dtype
                if type_map and key in type_map:
                    dtype = type_map[key].strip()

                    if dtype.lower() in data_types.REDSHIFT_MAP:
                        dtype = data_types.REDSHIFT_MAP[dtype.lower()]
                    elif dtype.upper() in data_types.REDSHIFT_ALIAS:
                        dtype = dtype.upper()

                # add mapping to key_map
                if mapped_key:
                    key_map[mapped_key] = ref_key
                    key = mapped_key
                else:
                    key = ref_key

                # determine dtype and generate new dict
                if isinstance(val, (dict, list)):
                    dtype = parse_types(val, parent=parent_key)
                    if dtype:
                        as_types[key] = dtype
                else:
                    inferred_dtype = data_types.set_dtype(val, infer_date=infer_date)

                    # detect dtype mismatch between inferred and user provided
                    if dtype and dtype != inferred_dtype:
                        user_dtype = dtype[:]
                        dtype = data_types.resolve_type(inferred_dtype, user_dtype)

                        if dtype != user_dtype:
                            logger.warning(
                                f"[{parent_key}] Using {dtype} instead of {user_dtype}"
                            )
                    else:
                        dtype = inferred_dtype

                    # skip keys with unknown data types and log
                    if "UNKNOWN" in dtype:
                        _str_dtype = dtype.split("_", 1)[1]
                        logger.warning(
                            f"Unknown dtype {_str_dtype} for {parent_key}: {val}"
                        )
                    else:
                        as_types[key] = dtype

        else:
            as_types = data_types.set_dtype(d, infer_date=infer_date)

        return as_types

    return parse_types(d), key_map


# --------------------------------------------------------------------------------------


def format_definitions(
    d,
    mapping=None,
    type_map=None,
    ignore_fields=None,
    infer_date=False,
    convert_hyphens=False,
    case_map=False,
    case_insensitive=False,
    ignore_nested_arrarys=True,
):
    """Format field names and set dtypes."""

    with_types, key_map = define_types(
        d,
        mapping=mapping,
        type_map=type_map,
        ignore_fields=ignore_fields,
        infer_date=infer_date,
        convert_hyphens=convert_hyphens,
        case_map=case_map,
        case_insensitive=case_insensitive,
        ignore_nested_arrarys=ignore_nested_arrarys,
    )

    if not with_types:
        logger.warning("Aborting - input does not contain valid data structures...")
        sys.exit(1)

    definitions = _conform_syntax(with_types)
    return definitions, key_map


def validate_input(d):
    """Check input type is dict|list."""

    if not d:
        raise ValueError("Input is empty...")

    if not isinstance(d, (dict, list)):
        raise ValueError("Invalid input type...")


def from_dict(
    d,
    mapping=None,
    type_map=None,
    ignore_fields=None,
    infer_date=False,
    convert_hyphens=False,
    schema=None,
    table=None,
    partitions=None,
    s3_key=None,
    case_map=False,
    case_insensitive=False,
    ignore_malformed_json=True,
    ignore_nested_arrarys=True,
    **kwargs,
):
    """Create Spectrum schema from dict."""

    validate_input(d)

    definitions, key_map = format_definitions(
        d,
        mapping=mapping,
        type_map=type_map,
        ignore_fields=ignore_fields,
        convert_hyphens=convert_hyphens,
        case_map=case_map,
        case_insensitive=case_insensitive,
        ignore_nested_arrarys=ignore_nested_arrarys,
    )

    statement = write_ddl.create_statement(
        definitions,
        key_map,
        schema,
        table,
        partitions,
        s3_key,
        case_insensitive,
        ignore_malformed_json,
    )

    return statement



# -*- coding: utf-8 -*-

import argparse
import io
import logging

from pathlib import Path
from typing import Dict, List

try:
    import ujson as json
except ImportError:
    import json

from . import __version__ as version
from . import ddl
from .MaxDict import MaxDict


logging.basicConfig(
    format="-- [%(asctime)s][%(levelname)s][%(name)s] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger = logging.getLogger(__name__)


def str_list_type(input):
    val = []
    if "," in input:
        val.extend((s.strip() for s in input.split(",") if s.strip()))
    else:
        val.append(input.strip())
    return val


def json_type(input: str):
    d = json.loads(Path(input))
    return tuple(d.items())


def parse_arguments():
    """Parse CLI args."""

    parser = argparse.ArgumentParser(
        description="Generate Athena and Spectrum DDL from JSON",
    )

    parser.set_defaults(partitions=None, mapping=None, type_map=None)

    parser.add_argument("-V", "--version", action="version", version=version)
    parser.add_argument(
        "-v", "--verbose", action="count", help="increase logging level"
    )

    parser.add_argument(
        dest="infile",
        type=argparse.FileType("r"),
        nargs="+",
        help="JSON files to convert",
    )

    case_group = parser.add_mutually_exclusive_group()

    case_group.add_argument(
        "-c",
        "--case_map",
        action="store_true",
        dest="case_map",
        help="disable case insensitivity and map field with uppercase chars to lowercase",
    )

    case_group.add_argument(
        "-l",
        "--lowercase",
        action="store_true",
        dest="case_insensitive",
        help="DDL: enable case insensitivity and force all fields to lowercase - applied before field lookup in mapping",
    )

    parser.add_argument(
        "-d",
        "--infer_date",
        action="store_true",
        help="infer date string types - supports ISO 8601 for date, datetime[TZ]",
    )

    parser.add_argument(
        "-r",
        "--retain_hyphens",
        action="store_false",
        dest="convert_hyphens",
        help="disable auto convert hypens to underscores",
    )

    parser.add_argument(
        "-e",
        "--error_nested_arrarys",
        action="store_false",
        dest="ignore_nested_arrarys",
        help="raise exception for nested arrays",
    )

    parser.add_argument(
        "-f",
        "--ignore_fields",
        type=str_list_type,
        dest="ignore_fields",
        metavar="col1,col2,...",
        help="Comma separated fields to ignore",
    )

    parser.add_argument(
        "-m",
        "--mapping",
        type=argparse.FileType("r"),
        dest="mapping_file",
        metavar="filepath",
        help="JSON filepath to use for mapping field names e.g. {field_name: new_field_name}",
    )
    parser.add_argument(
        "-y",
        "--type_map",
        type=argparse.FileType("r"),
        dest="type_map_file",
        metavar="filepath",
        help="JSON filepath to use for mapping field names to known data types e.g. {key: value}",
    )

    parser.add_argument(
        "-p",
        "--partitions_file",
        type=argparse.FileType("r"),
        dest="partitions_file",
        metavar="filepath",
        help="DDL: JSON filepath to map parition column(s) e.g. {column: dtype}",
    )

    parser.add_argument(
        "-j",
        "--ignore_malformed_json",
        action="store_true",
        dest="ignore_malformed_json",
        help="DDL: ignore malformed json",
    )
    parser.add_argument(
        "-s",
        "--schema",
        type=str,
        dest="schema",
        metavar="schema",
        help="DDL: schema name",
    )

    parser.add_argument(
        "-t",
        "--table",
        type=str,
        dest="table",
        metavar="table",
        help="DDL: table name",
    )

    parser.add_argument(
        "--s3",
        type=str,
        dest="s3_key",
        metavar="s3://bucket/key",
        help="DDL: S3 Key prefix",
    )

    args = parser.parse_args()

    if args.partitions_file:
        args.partitions = json.loads(args.partitions_file.read())

    if args.mapping_file:
        args.mapping = json.loads(args.mapping_file.read())

    if args.type_map_file:
        args.type_map = json.loads(args.type_map_file.read())

    if not args.verbose:
        logger.setLevel(logging.WARN)
    elif args.verbose == 1:
        logger.setLevel(logging.INFO)
    elif args.verbose > 1:
        logger.setLevel(logging.DEBUG)

    return args


def load_dict_from_infiles(infiles: List[io.TextIOWrapper]) -> Dict:
    """Load JSON infile(s) as dict | max dict.
    If > 1 files or any input is a list of dicts, max dict is used as schema source.
    """

    if len(infiles) == 1:
        d = json.loads(infiles[0].read())
        if isinstance(d, dict):
            return d
        else:
            infiles[0].seek(0)

    md = MaxDict()
    for infile in infiles:
        d = json.loads(infile.read())

        if isinstance(d, dict):
            md.load_dict(d)
        elif isinstance(d, list) and all(isinstance(sd, dict) for sd in d):
            md.batch_load_dicts(d)
        else:
            raise ValueError("Unsupported input, type must be dict or [dict,]...")

    return md.asdict()


def create_spectrum_schema():
    """Create Spectrum schema from JSON."""

    args = parse_arguments()
    d = load_dict_from_infiles(args.infile)

    kwargs = {k: v for (k, v) in args._get_kwargs()}
    statement = ddl.from_dict(d, **kwargs)
    print(statement)
